@@ -21,6 +21,7 @@
 
     Implemented features:
 
+        * Setting up with ``declarative.ext`` or without it.
         * Saving node roots --- if no parent set for node. The tree will have
           a new `tree_id`.
         * Saving child nodes --- if node has some parent. The whole dirty job
@@ -36,6 +37,9 @@
           basis of Adjacency Relations.
         * Collapsing flat tree returned from query to recursive structure (see
           :func:`tree_recursive_iterator`).
+        * Node classes may use `polymorphic inheritance
+          <http://www.sqlalchemy.org/docs/05/mappers.html
+          #mapping-class-inheritance-hierarchies>`_.
 
     Moving of nodes is not yet implemented.
 
@@ -55,16 +59,17 @@
 """
 import weakref
 from operator import attrgetter
-import sqlalchemy, sqlalchemy.orm
+import sqlalchemy, sqlalchemy.orm, sqlalchemy.orm.exc
+from sqlalchemy.orm.mapper import class_mapper
 from sqlalchemy.ext.declarative import DeclarativeMeta as BaseDeclarativeMeta
 
 
 __all__ = [
-    'MPManager', 'tree_recursive_iterator',
+    'MPManager', 'tree_recursive_iterator', 'DeclarativeMeta',
     'PathOverflowError', 'TooManyChildrenError', 'PathTooDeepError'
 ]
 
-__version__ = (0, 5)
+__version__ = (0, 5, 1)
 __doc__ %= {'version': '.'.join(map(str, __version__))}
 
 
@@ -157,7 +162,7 @@ class MPOptions(object):
 
         def _check_field(name, field, type_):
             # Check field argument (one of `path_field`, `depth_field` and
-            # `tree_id_field`),convert it from field name to `Column` object
+            # `tree_id_field`), convert it from field name to `Column` object
             # if needed, create the column object if needed and check the
             # existing `Column` object for sane.
             assert field
@@ -367,7 +372,6 @@ class MPClassManager(object):
     def __init__(self, node_class, opts):
         self._mp_opts = opts
         self.node_class = node_class
-        self.mapper_extension = MPMapperExtension(opts)
 
     @property
     def max_children(self):
@@ -500,12 +504,18 @@ class MPInstanceManager(object):
     Implements API to query nodes related somehow to particular
     node: descendants, ancestors, etc.
 
-    :param opts: instance of `MPOptions`
-    :param obj: particular node instance
+    :param opts:
+        instance of `MPOptions`.
+    :param root_node_class:
+        the root class in the node class' polymorphic inheritance hierarchy.
+        This class will be used to perform queries.
+    :param obj:
+        particular node instance.
     """
-    __slots__ = ('_mp_opts', '_obj_ref')
+    __slots__ = ('_mp_opts', '_obj_ref', '_root_node_class')
 
-    def __init__(self, opts, obj):
+    def __init__(self, opts, root_node_class, obj):
+        self._root_node_class = root_node_class
         self._mp_opts = opts
         self._obj_ref = weakref.ref(obj)
 
@@ -531,7 +541,7 @@ class MPInstanceManager(object):
             # use node's session only if particular session
             # was not specified
             session = obj_session
-        return sqlalchemy.orm.Query(type(obj), session=session)
+        return sqlalchemy.orm.Query(self._root_node_class, session=session)
 
     def _get_session_and_assert_flushed(self, obj):
         """
@@ -714,10 +724,14 @@ class MPManager(object):
 
     Now there is an ability to get instance manager or class manager
     via property `'mp'` depending on way to access it. `Node.mp` will
-    return class manager :class:`MPClassManager` for class `Node` and
-    `instance_node.mp` will return instance_node's
-    :class:`MPInstanceManager`. See that classes for more details about
-    their public API.
+    return mapper extension till class is mapped, class manager
+    :class:`MPClassManager` after that and `instance_node.mp`
+    will return instance_node's :class:`MPInstanceManager`.
+    See that classes for more details about their public API.
+
+    .. versionchanged:: 0.5.1
+        Previously mapper extension was accessible via class manager's
+        property.
 
     :param table:
         instance of `sqlalchemy.Table`. A table that will be mapped to
@@ -767,23 +781,62 @@ class MPManager(object):
         opts = MPOptions(*args, **kwargs)
         self._mp_opts = opts
         self.class_manager = None
+        self.mapper_extension = MPMapperExtension(opts=opts)
+        self.root_node_class = None
 
     def __get__(self, obj, objtype):
         """
-        Get :class:`MPInstanceManager` if was accessed via instance
-        or :class:`MPClassManager` if via class.
+        There may be three kinds of return values from this getter.
+
+        The first one is used when the class which this descriptor
+        is attached to is not yet mapped to any table. In that case
+        the return value is an instance of :class:`MPMapperExtension`.
+        which is intended to be used as mapper extension.
+
+        The second scenario is access to :class:`MPManager` via mapped
+        class. The corresponding :class:`MPInstanceManager`'s instance
+        is returned.
+
+        .. note:: If the nodes of your tree use polymorphic inheritance
+                  it is important to know that class manager is accessible
+                  only via the base class of inheritance hierarchy.
+
+        And the third way is accessing it from the node instance.
+        Attached to that node :class:`instance manager <MPInstanceManager>`
+        is returned then.
         """
         if obj is None:
+            try:
+                root_node_class = self.get_root_node_class(objtype)
+            except sqlalchemy.orm.exc.UnmappedClassError:
+                return self.mapper_extension
+            assert objtype is root_node_class, \
+                   "MPClassManager should be accessed via base class in the " \
+                   "polymorphic inheritance hierarchy: %r" % root_node_class
             if self.class_manager is None:
                 self.class_manager = MPClassManager(objtype, self._mp_opts)
             return self.class_manager
         else:
             instance_manager = obj.__dict__.get(self.instance_manager_key)
             if instance_manager is None:
-                instance_manager = MPInstanceManager(self._mp_opts, obj)
+                root_node_class = self.get_root_node_class(objtype)
+                instance_manager = MPInstanceManager(
+                    self._mp_opts, root_node_class, obj
+                )
                 obj.__dict__[self.instance_manager_key] = instance_manager
             return instance_manager
 
+    def get_root_node_class(self, objtype):
+        """
+        Get the root node class in the polymorphic inheritance hierarchy.
+        """
+        if self.root_node_class is None:
+            mapper = class_mapper(objtype)
+            while mapper.inherits is not None:
+                mapper = mapper.inherits
+            self.root_node_class = mapper.class_
+        return self.root_node_class
+
 
 _nonexistent = object()
 def _iter_current_next(sequence):
@@ -924,6 +977,10 @@ class DeclarativeMeta(BaseDeclarativeMeta):
             super(DeclarativeMeta, cls).__init__(name, bases, dct)
             return
         mp_manager_name = cls.__mp_manager__
+
+        # preventing the property to be inherited
+        del cls.__mp_manager__
+
         opts = {}
         for opt, default in [
                 ('path_field', 'mp_path'),
@@ -932,7 +989,13 @@ class DeclarativeMeta(BaseDeclarativeMeta):
                 ('steplen', 3),
                 ('instance_manager_key', '_mp_instance_manager')
             ]:
-            opts[opt] = getattr(cls, '__mp_%s__' % opt, default)
+            optname = '__mp_%s__' % opt
+            if hasattr(cls, optname):
+                opts[opt] = getattr(cls, optname)
+                delattr(cls, optname)
+            else:
+                opts[opt] = default
+
         for field, ftype in [
                 ('path_field', PathField),
                 ('depth_field', DepthField),
@@ -945,5 +1008,5 @@ class DeclarativeMeta(BaseDeclarativeMeta):
         mp_manager = MPManager(cls.__table__, **opts)
         setattr(cls, mp_manager_name, mp_manager)
         mp_class_manager = getattr(cls, mp_manager_name)
-        cls.__mapper__.extension.append(mp_class_manager.mapper_extension)
+        cls.__mapper__.extension.append(mp_manager.mapper_extension)
 
