@@ -63,7 +63,7 @@ def setup():
 
 
     sqlalchemy.orm.mapper(
-        Cls, tbl, extension=[Cls.mp.mapper_extension],
+        Cls, tbl, extension=[Cls.mp],
         properties={
             'parent': sqlalchemy.orm.relation(Cls, remote_side=[tbl.c.id])
         }
